import codecs
from os import path

testdata_dir = path.join(path.dirname(__file__), 'test_api_responses')


def load_data(filename):
    full_path = path.join(testdata_dir, filename)
    with codecs.open(full_path, encoding="utf-8") as f:
        return f.read()



import re
import warnings
from string import Template


def format_phone(phone):
    warnings.warn(
        'The phone number formatter will be deprecated in a future release of '
        'govdelivery. Going forward, please ensure that your application '
        'passes phone numbers in the format that the GovDelivery API requires '
        '(a ten-digit string with no punctuation).',
        DeprecationWarning
    )

    # if phone starts with + or 1, strip those characters
    phone = phone.lstrip('+1')
    # strip all other non-digit characters
    phone = re.sub(r'\D', '', phone)

    return phone


def create_subscriber(contact_details,
                      contact_method='email',
                      send_notifications=False,
                      digest_for=0):
    send_notifications = str(send_notifications).lower()

    country_code = ''
    if contact_method == 'phone':
        # Note: Currently only supports U.S. phone numbers
        contact_details = format_phone(contact_details)
        country_code = '<country-code>1</country-code>'

    subscriber_template = """<subscriber>
    <$contact_method>$contact_details</$contact_method>
    $country_code
    <send-notifications type='boolean'>$send_notifications</send-notifications>
    <digest-for>$digest_for</digest-for>
</subscriber>"""

    template = Template(subscriber_template)

    return template.substitute(locals())


def set_subscriber_categories(codes, send_notifications=False):
    send_notifications = str(send_notifications).lower()

    categories_xml = ''
    for code in codes:
        categories_xml += """
        <category>
            <code>%s</code>
        </category>
""" % code

    xml_template = """<subscriber>
    <send-notifications type='boolean'>$send_notifications</send-notifications>
    <categories type='array'>
        $categories_xml
    </categories>
</subscriber>"""

    template = Template(xml_template)
    return template.substitute(locals())


def set_subscriber_topics(codes,
                          contact_details,
                          contact_method='email',
                          send_notifications=False):
    send_notifications = str(send_notifications).lower()

    country_code = ''
    if contact_method == 'phone':
        # Note: Currently only supports U.S. phone numbers
        contact_details = format_phone(contact_details)
        country_code = '<country-code>1</country-code>'

    topics_xml = ''
    for code in codes:
        topics_xml += """
        <topic>
            <code>%s</code>
        </topic>
""" % code

    xml_template = """<subscriber>
    <$contact_method>$contact_details</$contact_method>
    $country_code
    <send-notifications type='boolean'>$send_notifications</send-notifications>
    <topics type='array'>
        $topics_xml
    </topics>
</subscriber>"""

    template = Template(xml_template)
    return template.substitute(locals())


def free_response_to_question(question_id, answer_text):
    xml_template = """<responses type="array">
    <response>
        <question-answer-text>$answer_text</question-answer-text>
        <question-id>$question_id</question-id>
        <answer-id nil="true"></answer-id>
    </response>
</responses>
"""

    template = Template(xml_template)
    return template.substitute(locals())


def select_response_to_question(question_id, answer_id):
    xml_template = """<responses type="array">
    <response>
        <question-id>$question_id</question-id>
        <answer-id>$answer_id</answer-id>
    </response>
</responses>
"""

    template = Template(xml_template)
    return template.substitute(locals())



import base64
import os
import urllib.parse as urlparse
from string import Template

import requests

from govdelivery import xml_payloads, xml_response_parsers

GOVDELIVERY_BASE_URL = 'https://api.govdelivery.com'


def get_full_url_to_call(path, base_url=None):
    return urlparse.urljoin(
        base_url or os.getenv('GOVDELIVERY_BASE_URL') or GOVDELIVERY_BASE_URL,
        path
    )


def authenticated_session(username=None, password=None):
    """
    Returns a request.Session with the Authorization
    headers set.
    """

    # allow credentials to come from os.environ
    # TODO: A sensible exception for when the variables don't exist
    # and haven't been passed to the function

    username = username or os.environ['GOVDELIVERY_USER']
    password = password or os.environ['GOVDELIVERY_PASSWORD']

    session = requests.Session()
    session.auth = (username, password)

    return session


def authenticated_api_call(path, method, payload=None, session=None,
                           base_url=None):
    """
    makes an authenticated call to the govdelivery API
    using the provided path, method, payload, and request.Session

    Returns a request.Response
    """

    url_to_call = get_full_url_to_call(path, base_url)

    session = session or authenticated_session()
    if method.lower() in ['post', 'put']:
        session.headers['content-type'] = 'application/XML'
    client = getattr(session, method.lower())

    response = client(url_to_call, data=payload)
    return response


class GovDelivery(object):
    """
    Higher-level wrapper for the GovDelivery API
    """

    def __init__(self, username=None, password=None, account_code=None,
                 base_url=None):

        self.session = authenticated_session(username, password)
        self.account_code = (
            account_code or os.environ['GOVDELIVERY_ACCOUNT_CODE']
        )

    # TODO: Replace this function with f-strings.
    def translate_path(self, path_template, **kwargs):
        context = kwargs
        context['account_code'] = self.account_code
        template = Template(path_template)
        return template.substitute(**kwargs)

    def call_api(self, path, method='get', payload=None, response_parser=None):
        unparsed_response = authenticated_api_call(
            path,
            method,
            payload,
            self.session
        )

        if not response_parser:
            return unparsed_response

        return response_parser(unparsed_response.text)

    def create_subscriber(self, contact_details, contact_method='email',
                          send_notifications=False, digest_for=0):
        payload = xml_payloads.create_subscriber(
            contact_details,
            contact_method,
            send_notifications,
            digest_for
        )

        path = self.translate_path(
            '/api/account/$account_code/subscribers.xml'
        )
        return self.call_api(path, 'post', payload)

    def get_all_topics(self):
        path = self.translate_path('/api/account/$account_code/topics.xml')
        return self.call_api(
            path,
            'get',
            response_parser=xml_response_parsers.topic_xml_as_dict
        )

    # TODO: Rename method to be more precise
    def get_visible_topics(self):
        path = self.translate_path('/api/account/$account_code/topics.xml')
        return self.call_api(
            path,
            'get',
            response_parser=xml_response_parsers.listed_topic_xml_as_dict
        )

    def get_subscriber_categories(self, contact_details):
        subscriber_id = base64.b64encode(contact_details.encode('utf-8'))
        path = self.translate_path(
            '/api/account/$account_code/'
            'subscribers/$subscriber_id/'
            'categories.xml',
            subscriber_id=subscriber_id
        )
        return self.call_api(path)

    def set_subscriber_categories(self, contact_details, category_codes,
                                  send_notifications=False):
        subscriber_id = base64.b64encode(contact_details.encode('utf-8'))
        path = self.translate_path(
            '/api/account/$account_code/'
            'subscribers/$subscriber_id/'
            'categories.xml',
            subscriber_id=subscriber_id
        )
        payload = xml_payloads.set_subscriber_categories(
            category_codes,
            send_notifications
        )
        return self.call_api(path, 'put', payload)

    def get_subscriber_topics(self, contact_details):
        subscriber_id = base64.b64encode(contact_details.encode('utf-8'))
        path = self.translate_path(
            '/api/account/$account_code/subscribers/$subscriber_id/topics.xml',
            subscriber_id=subscriber_id
        )
        return self.call_api(
            path,
            response_parser=xml_response_parsers.subscriber_topics_as_list
        )

    def set_subscriber_topics(self, contact_details, topic_codes,
                              contact_method='email',
                              send_notifications=False):
        topic_code_set = set(topic_codes)

        path = self.translate_path(
            '/api/account/$account_code/subscriptions.xml'
        )
        payload = xml_payloads.set_subscriber_topics(
            topic_code_set,
            contact_details,
            contact_method,
            send_notifications
        )
        return self.call_api(path, 'post', payload)

    def set_subscriber_answers_to_question(self, contact_details, question_id,
                                           answer_text):
        subscriber_id = base64.b64encode(contact_details.encode('utf-8'))
        question_id_encoded = base64.b64encode(question_id.encode('utf-8'))

        path = self.translate_path(
            '/api/account/$account_code/'
            'subscribers/$subscriber_id/'
            'questions/$question_id_encoded/'
            'responses.xml',
            subscriber_id=subscriber_id,
            question_id_encoded=question_id_encoded
        )

        payload = xml_payloads.free_response_to_question(
            question_id_encoded,
            answer_text
        )
        return self.call_api(path, 'put', payload)

    def set_subscriber_answer_to_select_question(
        self,
        contact_details,
        question_id,
        answer_id
    ):
        subscriber_id = base64.b64encode(contact_details.encode('utf-8'))
        question_id_encoded = base64.b64encode(question_id.encode('utf-8'))
        answer_id_encoded = base64.b64encode(answer_id.encode('utf-8'))

        path = self.translate_path(
            (
                '/api'
                '/account/$account_code'
                '/subscribers/$subscriber_id'
                '/questions/$question_id_encoded'
                '/responses.xml'
            ),
            subscriber_id=subscriber_id,
            question_id_encoded=question_id_encoded,
            answer_id_encoded=answer_id_encoded
        )

        payload = xml_payloads.select_response_to_question(
            question_id_encoded,
            answer_id_encoded
        )
        return self.call_api(path, 'put', payload)



import unittest

from govdelivery.tests.utils import load_data
from govdelivery.xml_response_parsers import (
    listed_topic_xml_as_dict,
    subscriber_responses_as_list_of_dicts,
    subscriber_topics_as_list,
    topic_xml_as_dict,
)


class TestTopicListingParsers(unittest.TestCase):
    def test_topic_xml_as_dict(self):
        response_xml = load_data('list_all_topics.xml')
        parsed = topic_xml_as_dict(response_xml)
        self.assertEqual(
            parsed,
            {'XYZ_998': 'Test Unlisted Topic', 'XYZ_999': 'Test Topic'}
        )

    def test_listed_topic_xml_as_dict(self):
        response_xml = load_data('list_all_topics.xml')
        parsed = listed_topic_xml_as_dict(response_xml)
        self.assertEqual(
            parsed,
            {'XYZ_999': 'Test Topic'}
        )


class TestSubscriberTopicsParser(unittest.TestCase):
    def test_subscriber_topics_as_list(self):
        response_xml = load_data('list_subscriber_topics.xml')
        parsed = subscriber_topics_as_list(response_xml)
        self.assertEqual(
            parsed,
            ['XYZ_999']
        )


class TestQuestionResponseParser(unittest.TestCase):
    def test_parse_response(self):
        response_xml = load_data('question_responses.xml')
        parsed = subscriber_responses_as_list_of_dicts(response_xml)
        r1, r2, r3 = parsed

        self.assertIn('question_id', r1)
        self.assertEqual(r1['question_id'], 'MTAwNDk=')



import xml.etree.ElementTree as ET


def topic_xml_as_dict(unparsed_xml, only_listed=False):
    root = ET.fromstring(unparsed_xml)
    code_map = {}

    for topic_tag in root.findall('topic'):
        if topic_tag.find('visibility').text == 'Unlisted' and only_listed:
            continue
        name = topic_tag.find('name').text
        code = topic_tag.find('code').text
        code_map[code] = name

    return code_map


def listed_topic_xml_as_dict(unparsed_xml):
    return topic_xml_as_dict(unparsed_xml, only_listed=True)


def subscriber_topics_as_list(unparsed_xml):
    root = ET.fromstring(unparsed_xml)
    codes = []

    for topic in root.findall('topic'):
        codes.append(topic.find('to-param').text)

    return codes


# This parser is currently unused.
# TODO: Write corresponding get_subscriber_answers client method.
def subscriber_responses_as_list_of_dicts(unparsed_xml):
    root = ET.fromstring(unparsed_xml)
    responses = []

    for response_tag in root.findall('response'):
        answer_id_tag = response_tag.find('answer-id')

        question_id = response_tag.find('question-id').text

        answer_id = answer_id_tag.text
        answer_id_is_null = answer_id_tag.attrib.get('nil') == 'true'

        answer_text = response_tag.find('question-answer-text').text

        response_dict = {
            'question_id': question_id,
            'answer_id': answer_id,
            'answer_text': answer_text,
            'answer_id_is_null': answer_id_is_null,
        }

        responses.append(response_dict)

    return responses



import unittest

from govdelivery.xml_payloads import (
    create_subscriber,
    format_phone,
    set_subscriber_categories,
    set_subscriber_topics,
)


class TestXMLPayloadUtils(unittest.TestCase):
    def test_format_phone_removes_plus_one_and_all_non_digits(self):
        phone = '+1 (234) 567-8901'
        self.assertEqual('2345678901', format_phone(phone))

    def test_format_phone_removes_leading_one_and_all_non_digits(self):
        phone = '1 (234) 567-8901'
        self.assertEqual('2345678901', format_phone(phone))


class TestXMLPayloadGenerators(unittest.TestCase):
    def test_create_subscriber_with_email(self):
        contact_details = 'mail@example.com'
        payload = create_subscriber(contact_details)
        self.assertIn('<email>mail@example.com</email>', payload)

    def test_create_subscriber_with_phone(self):
        contact_details = '2345678901'
        contact_method = 'phone'
        payload = create_subscriber(contact_details, contact_method)
        self.assertIn('<phone>2345678901</phone>', payload)

    def test_set_subscriber_categories(self):
        codes = ['XYZ_C1', 'XYZ_C2']
        payload = set_subscriber_categories(codes)
        self.assertIn('<code>XYZ_C1</code>', payload)
        self.assertIn('<code>XYZ_C2</code>', payload)

    def test_set_subscriber_topics_with_email(self):
        contact_details = 'mail@example.com'
        codes = ['XYZ_1', 'XYZ_2']
        payload = set_subscriber_topics(codes, contact_details)
        self.assertIn('<email>mail@example.com</email>', payload)
        self.assertIn('<code>XYZ_1</code>', payload)
        self.assertIn('<code>XYZ_2</code>', payload)

    def test_set_subscriber_topics_with_phone(self):
        contact_details = '2345678901'
        contact_method = 'phone'
        codes = ['XYZ_1', 'XYZ_2']
        payload = set_subscriber_topics(codes, contact_details, contact_method)
        self.assertIn('<phone>2345678901</phone>', payload)
        self.assertIn('<code>XYZ_1</code>', payload)
        self.assertIn('<code>XYZ_2</code>', payload)



import unittest

from govdelivery.tests.utils import load_data


class TestLoadData(unittest.TestCase):
    def test_load_data(self):
        self.assertEqual(
            load_data('list_subscriber_topics.xml'),
            """<?xml version="1.0" encoding="UTF-8"?>
<topics type="array">
    <topic>
        <to-param>XYZ_999</to-param>
        <link rel="self" href="/api/account/XYZ/topics/XYZ_999"/>
        <topic-uri>/api/account/XYZ/topics/XYZ_999.xml</topic-uri>
    </topic>
</topics>
"""
        )



from setuptools import find_packages, setup


long_description = open('README.md', 'r').read()

install_requires = ['requests>=2.22.0,<3']

testing_extras = [
    'coverage',
    'responses',
]

setup(
    name='govdelivery',
    url='https://github.com/cfpb/govdelivery',
    author='CFPB',
    author_email='tech@cfpb.gov',
    description='Python module for interacting with the GovDelivery API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='CC0',
    version='1.4.0',
    include_package_data=True,
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        'testing': testing_extras,
    },
    test_suite='govdelivery.tests',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications :: Email',
    ],
    zip_safe=False,
)



import base64
import unittest
from unittest.mock import patch

import responses
from requests.models import Response

from govdelivery.api import (
    GovDelivery,
    authenticated_session,
    get_full_url_to_call,
)
from govdelivery.tests.utils import load_data
from govdelivery.xml_response_parsers import topic_xml_as_dict


class TestClientUtilities(unittest.TestCase):
    @patch('os.getenv', return_value=None)
    def test_get_full_url_to_call_with_default_base_url(self, _):
        self.assertEqual(
            get_full_url_to_call('/foo/bar'),
            'https://api.govdelivery.com/foo/bar'
        )

    @patch('os.getenv', return_value='https://envvar.govdelivery.com')
    def test_get_full_url_to_call_with_env_var_defined_base_url(self, _):
        self.assertEqual(
            get_full_url_to_call('/foo/bar'),
            'https://envvar.govdelivery.com/foo/bar'
        )

    def test_get_full_url_to_call_with_passed_base_url(self):
        self.assertEqual(
            get_full_url_to_call(
                '/foo/bar',
                base_url='https://passed.govdelivery.com'
            ),
            'https://passed.govdelivery.com/foo/bar'
        )

    def test_authenticated_session(self):
        session = authenticated_session('username', 'password')
        self.assertEqual(session.auth, ('username', 'password'))

    def test_authenticated_api_call(self):
        # There is no meaningful way to directly test this method, and it
        # should get covered anyway by testing the other methods that call it.
        pass


class TestGovDelivery(unittest.TestCase):
    def setUp(self):
        self.gd = GovDelivery(
            username='abc@xyz.123',
            password='mycatsname',
            account_code='XYZ'
        )

    def test_translate_path(self):
        path = self.gd.translate_path('/api/account/$account_code/topics.xml')
        self.assertEqual(path, '/api/account/XYZ/topics.xml')

    @responses.activate
    def test_call_api_no_parser_returns_unparsed(self):
        path = self.gd.translate_path('/api/account/$account_code/topics.xml')
        mockResponseBody = load_data('list_all_topics.xml')
        responses.add(
            responses.GET,
            get_full_url_to_call(path),
            body=mockResponseBody,
            status=200
        )
        response = self.gd.call_api(path, 'get')
        self.assertIsInstance(response, Response)

    @responses.activate
    def test_call_api_with_parser_returns_parsed(self):
        path = self.gd.translate_path('/api/account/$account_code/topics.xml')
        mockResponseBody = load_data('list_all_topics.xml')
        responses.add(
            responses.GET,
            get_full_url_to_call(path),
            body=mockResponseBody,
            status=200
        )
        parsed_response = self.gd.call_api(
            path,
            'get',
            response_parser=topic_xml_as_dict
        )
        self.assertEqual(
            parsed_response,
            {'XYZ_998': 'Test Unlisted Topic', 'XYZ_999': 'Test Topic'}
        )

    @responses.activate
    def test_get_all_topics(self):
        path = self.gd.translate_path('/api/account/$account_code/topics.xml')
        mockResponseBody = load_data('list_all_topics.xml')
        responses.add(
            responses.GET,
            get_full_url_to_call(path),
            body=mockResponseBody,
            status=200
        )
        response = self.gd.get_all_topics()
        self.assertEqual(
            response,
            {'XYZ_998': 'Test Unlisted Topic', 'XYZ_999': 'Test Topic'}
        )

    @responses.activate
    def test_get_visible_topics(self):
        path = self.gd.translate_path('/api/account/$account_code/topics.xml')
        mockResponseBody = load_data('list_all_topics.xml')
        responses.add(
            responses.GET,
            get_full_url_to_call(path),
            body=mockResponseBody,
            status=200
        )
        response = self.gd.get_visible_topics()
        self.assertEqual(response, {'XYZ_999': 'Test Topic'})

    @responses.activate
    def test_get_subscriber_topics(self):
        email = 'test@example.com'
        subscriber_id = base64.b64encode(email.encode('utf-8'))
        path = self.gd.translate_path(
            '/api/account/$account_code/subscribers/$subscriber_id/topics.xml',
            subscriber_id=subscriber_id
        )
        mockResponseBody = load_data('list_subscriber_topics.xml')
        responses.add(
            responses.GET,
            get_full_url_to_call(path),
            body=mockResponseBody,
            status=200
        )
        response = self.gd.get_subscriber_topics('test@example.com')
        self.assertIn('XYZ_999', response)
